/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.army.hierarchy;

import java.awt.Color;
import java.awt.Image;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;

import fb.main.LoadingException;
import fb.util.FBColors;
import fb.util.Resources;
import fb.util.StringTools;

/**
 * a campaign force
 * 
 * @author farfpeuf
 * 
 */
public class Force extends HierarchyNode {
	// separator tag
	public static final String FORCE_TAG = "#force#";

	// property change constants
	public static final String BUDGET_PROPERTY = "budget";
/**
 * icon used in tree table
 */
	private Icon minicon;
	
	/**
	 * player name and force race
	 */
	private String playerName, race;

	/**
	 * lists
	 */
	private Vector armies = new Vector(), ennemies = new Vector(),
			friends = new Vector();

	/**
	 * artificial intelligence or human player ?
	 */
	private boolean ai = false;

	/**
	 * handled by this computer ?
	 */
	private boolean handled = false;

	/**
	 * id
	 */
	private int id;

	/**
	 * side
	 */
	private int side;

	/**
	 * budget
	 */
	private int budget;
/**
 * constructor
 * @param playerName player to which this force belongs
 * @param race force race
 * @param isAI artificial intelligence or human player ?
 * @param side side
 * @param isHandled handled by this computer ?
 * @param budget initial budget
 */
	public Force(String playerName, String race, boolean isAI, int side,
			boolean isHandled, int budget) {
		this.playerName = playerName;
		this.race = race;
		this.ai = isAI;
		this.side = side;
		this.handled = isHandled;
		this.budget = budget;
	}
/**
 * receive campaign income
 * @param income
 */
	public void receiveIncome(int income) {
		int oldBudget = budget;
		budget += income;
		firePropertyChange(BUDGET_PROPERTY, oldBudget, budget);
	}
/**
 * calculate cost
 */
	public int getCost() {
		int cost = 0;
		for (int i = 0; i < armies.size(); i++) {
			cost += ((Army) armies.get(i)).getCost();
		}
		return cost;
	}
/**
 * @return number of warriors
 */
	public int getNumberOfWarriors() {
		int number = 0;
		for (int i = 0; i < armies.size(); i++) {
			number += ((Army) armies.get(i)).getNumberOfWarriors();
		}
		return number;
	}
	/**
	 * calculate force degradation
	 * 
	 * @return force degradation
	 */
	public int degradation() {
		int degradation = 0;

		for (int i = 0; i < armies.size(); i++) {
			degradation += ((Army) armies.get(i)).degradation();
		}

		return degradation;
	}
	/**
	 * add a ennemy to this force
	 * 
	 * @param ennemy
	 *            the ennemy force
	 */
	public void addEnnemy(Force ennemy) {
		ennemies.add(ennemy);
	}
	/**
	 * add a friendly force to this army
	 * 
	 * @param ennemy
	 *            the friendly force
	 */
	public void addFriend(Force friend) {
		friends.add(friend);
	}
/**
 * add an army to this force
 * @param army the new army
 */
	public void addArm(Army army) {
		armies.add(army);
		army.setForce(this);

	}
/**
 * remove an army from this force
 * @param army the army to remove
 */
	public void removeArmy(Army army) {
		armies.remove(army);
		army.setForce(null);
	}

	/**
	 * Sets the race.
	 * 
	 * @param race
	 *            The race to set
	 */
	public void setRace(String race) {
		this.race = race;
	}

	/**
	 * Returns the id.
	 * 
	 * @return int
	 */
	public int getId() {
		return id;
	}

	/**
	 * Sets the id.
	 * 
	 * @param id
	 *            The id to set
	 */
	public void setId(int id) {
		this.id = id;
	}
	/**
	 * returns force as a string to save it to a file
	 * 
	 * @return force as a string
	 */
	public String saveBattleString() {
		String saveString = race + '\t' + playerName + '\t' + id + '\t' + side
				+ '\t' + (ai ? 'y' : 'n') + '\t' + (handled ? 'y' : 'n') + '\t'
				+ budget + '\t';
		for (int i = 0; i < armies.size(); i++) {
			saveString += Army.ARM_TAG
					+ (((Army) armies.get(i)).saveBattleString());
		} // for
		return saveString;
	}
/**
 * constructor from string read in a file
 * @param fileString string read in a file
 * @throws LoadingException
 */
	public Force(String fileString) throws LoadingException {
		String[] tokens = fileString.split(Army.ARM_TAG);

		StringBuffer buffy = new StringBuffer(tokens[0]);

		race = StringTools.readBuffer(buffy);
		playerName = StringTools.readBuffer(buffy);
		id = StringTools.readBufferInt(buffy);
		side = StringTools.readBufferInt(buffy);
		ai = (StringTools.readBufferChar(buffy) == 'y') ? true : false;
		handled = (StringTools.readBufferChar(buffy) == 'y') ? true : false;
		budget = StringTools.readBufferInt(buffy);

		Army arm;
		for (int i = 1; i < tokens.length; i++) {
			arm = new Army(tokens[i]);
			addArm(arm);
		}
	}

	/**
	 * @see javax.swing.tree.TreeNode#children()
	 */
	public Enumeration children() {
		return armies.elements();
	}

	/**
	 * @see javax.swing.tree.TreeNode#getAllowsChildren()
	 */
	public boolean getAllowsChildren() {
		return true;
	}

	/**
	 * @see javax.swing.tree.TreeNode#getChildAt(int)
	 */
	public TreeNode getChildAt(int childIndex) {
		return (TreeNode) armies.get(childIndex);
	}

	/**
	 * @see javax.swing.tree.TreeNode#getChildCount()
	 */
	public int getChildCount() {
		return armies.size();
	}

	/**
	 * @see javax.swing.tree.TreeNode#getIndex(TreeNode)
	 */
	public int getIndex(TreeNode node) {
		return armies.indexOf(node);
	}

	/**
	 * @see javax.swing.tree.TreeNode#getParent()
	 */
	public TreeNode getParent() {
		return null;
	}

	/**
	 * @see javax.swing.tree.TreeNode#isLeaf()
	 */
	public boolean isLeaf() {
		return false;
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#insert(MutableTreeNode, int)
	 */
	public void insert(MutableTreeNode child, int index) {
		addArm((Army) child);
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#remove(int)
	 */
	public void remove(int index) {
		removeArmy((Army) armies.get(index));
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#remove(MutableTreeNode)
	 */
	public void remove(MutableTreeNode node) {
		removeArmy((Army) node);
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#removeFromParent()
	 */
	public void removeFromParent() {
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#setParent(MutableTreeNode)
	 */
	public void setParent(MutableTreeNode newParent) {
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#setUserObject(Object)
	 */
	public void setUserObject(Object object) {
	}
/**
 * @return force full name
 */
	public String getFullName() {
		return playerName;
	}


/**
 * @return icon used in tree table
 */
	public Icon getMinicon() {
		if (null == minicon) {
			minicon = new ImageIcon((Resources.getSign(Resources
					.getString(StringTools.replaceAll(getRace(), " ", ""))))
					.getImage().getScaledInstance(20, 20, Image.SCALE_SMOOTH));
		}
		return minicon;
	}
	/**
	 * return number of entities of specified class
	 * 
	 * @param cl
	 *            class to search for
	 * @return number of entities of specified class
	 */
	public int numberOf(Class cl) {
		int number = 0;
		Iterator iter = armies.iterator();
		while (iter.hasNext()) {
			number += ((Army) iter.next()).numberOf(cl);
		}
		return number;
	}
/**
 * @return the force strongest army
 */
	public Army getStrongestArm() {
		Army strongest = null, tmp;
		Iterator iter = armies.iterator();
		while (iter.hasNext()) {
			tmp = (Army) iter.next();
			if (strongest == null || tmp.getCost() > strongest.getCost()) {
				strongest = tmp;
			}
		}
		return strongest;
	}



	/**
	 * @return Returns the budget.
	 */
	public int getBudget() {
		return budget;
	}
	/**
	 * @param budget The budget to set.
	 */
	public void setBudget(int budget) {
		this.budget = budget;
	}
	/**
	 * @return Returns the armies.
	 */
	public Vector getArmies() {
		return armies;
	}
	/**
	 * @return Returns the playerName.
	 */
	public String getPlayerName() {
		return playerName;
	}
	/**
	 * @return Returns the race.
	 */
	public String getRace() {
		return race;
	}
	/**
	 * @return the force color
	 */
	public Color getColor() {
		return FBColors.color[id];
	}
	/**
	 * @return Returns the ai.
	 */
	public boolean isAi() {
		return ai;
	}
	/**
	 * @return Returns the handled.
	 */
	public boolean isHandled() {
		return handled;
	}
	/**
	 * @return Returns the side.
	 */
	public int getSide() {
		return side;
	}




}