/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.battlefieldelements;

import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;

import javax.swing.ImageIcon;

import fb.army.entities.Entity;
import fb.battlefield.BattleFieldModel;

/**
 * a sign painted on the battlefield
 * 
 * @author farfpeuf
 * 
 */
public class Sign extends Animated {
	// current image
	private ImageIcon image = null;

	// current position
	protected int x = 0, y = 0;

	// offsets
	private int signDx, signDy;

	// size
	private int width, length;

	// destination
	private int xd, yd;

	// current action
	private char action = 's';

	// step
	private int step;

	/**
	 * create a sign attached to a specified entity
	 * @param e the entity to which the sign is attached
	 */
	public Sign(Entity e) {
		step = (int) (e.getMovement() / 2);
		BattleFieldModel.getInstance().addSign(this);
		width = 8;
		length = 8;
		signDx = (int) (0 - width / 2);
		signDy = (int) (0 - length / 2);
	}

	/**
	 * teleport the sign to its destination
	 *
	 */
	public void teleport() {
		doMove(xd - x, yd - y);
	}

	/**
	 * change sign image
	 * @param image the new image
	 */
	public void setImage(ImageIcon image) {
		teleport();
		if (this.image != image) {
			ImageIcon oldImage = this.image;
			this.image = image;
			firePropertyChange(IMAGE, oldImage, image);
		}
	}

	/**
	 * clear sign image
	 *
	 */
	public void clear() {
		image = null;
	}

	/**
	 * move the sign to a new destianation
	 * @param nx new x coordinate
	 * @param ny new y coordinate
	 */
	public void moveto(int nx, int ny) {
		action = 'w';
		xd = nx;
		yd = ny;
	}

	/**
	 * rotate the sign
	 * @param theta rotation angle
	 * @param xc rotation center x coordinate
	 * @param yc rotation center y coordinate
	 */
	public void rotate(double theta, int xc, int yc) {
		int xdest = xd;
		int ydest = yd;
		xdest -= xc;
		ydest -= yc;
		Point2D p = AffineTransform.getRotateInstance(-theta).deltaTransform(
				new Point(xdest, ydest), null);
		xdest = (int) p.getX();
		ydest = (int) p.getY();
		xd = xdest + xc;
		yd = ydest + yc;
		action = 'w';
	}

	/**
	 * get sign bounds
	 */
	public Rectangle getBounds() {
		return new Rectangle(x + signDx, y + signDy, width, length);
	}

	/**
	 * move sign to destination
	 */
	public void animate() {
		if (image != null) {
			if (action == 'w') {
				int distx, disty, pasx, pasy;

				distx = Math.abs(xd - x);
				disty = Math.abs(yd - y);
				if (distx > disty) {
					pasx = (int) ((double) step / (double) (1.0 + (double) disty
							/ (double) distx));
					pasy = step - pasx;
				} else {
					pasy = (int) ((double) step / (double) (1.0 + (double) distx
							/ (double) disty));
					pasx = step - pasy;
				}
				if (pasx > distx) {
					pasx = distx;
				}
				if (pasy > disty) {
					pasy = disty;
				}
				if (x > xd) {
					pasx = -pasx;
				}
				if (y > yd) {
					pasy = -pasy;
				}
				doMove(pasx, pasy);

				if (x == xd && y == yd) { // fin de marche
					action = 's';
				}
			}
		}
	}
	/**
	 * apply specified move
	 * 
	 * @param xOffset
	 *            x Offset
	 * @param yOffset
	 *            y Offset
	 */
	public void doMove(int xOffset, int yOffset) {
		if (xOffset != 0 || yOffset != 0) {
			Rectangle previousBounds = getBounds();
			x += xOffset;
			y += yOffset;
			firePropertyChange(BOUNDS, previousBounds, getBounds());
		}
	}



	/**
	 * @return Returns the x.
	 */
	public int getX() {
		return x;
	}

	/**
	 * @return Returns the y.
	 */
	public int getY() {
		return y;
	}

	/**
	 * paint the sign on the battlefield
	 */
	public void paintImage(Graphics g, int dx, int dy, int scale) {
		if (image != null) {
			g.drawImage(image.getImage(), (x + signDx) * scale + dx,
					(y + signDy) * scale + dy, width * scale, length * scale,
					null);
		}
	}
}