/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.battlefieldelements;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.List;

import fb.army.entities.Entity;
import fb.army.entities.FlyingEntity;
import fb.battlefield.BattleFieldModel;
import fb.main.FB;
import fb.network.Client;
import fb.network.MessagesConstants;
import fb.util.DX;
import fb.util.GeometryTools;

/**
 * a warrior (either an individual or an element of a unit)
 * 
 * @author farfpeuf
 * 
 */
public class Warrior extends Figure implements Cloneable,
		PropertyChangeListener {

	// fighting contact areas
	private Area contactArea, spearContactArea, pikeContactArea;

	// altitude (0 for ground units)
	private int altitude;

	// must the target die ?
	private boolean mustDie = false;

	// life points
	private int lifePoints;

	// entity to which this warrior belongs
	private Entity entity;

	private List friendlycontacts = new ArrayList(),
			ennemyContacts = new ArrayList(),
			ennemySpearContacts = new ArrayList(),
			ennemyPikeContacts = new ArrayList();

	/**
	 * constructor
	 * 
	 * @param entity
	 *            entity to which this warrio is attaches
	 * @param lifePoints
	 *            initial life points
	 */
	public Warrior(Entity entity, int lifePoints) {
		super(entity);
		this.entity = entity;
		entity.addPropertyChangeListener(Entity.SCALE, this);
		entity.addPropertyChangeListener(Entity.FLYING, this);
		altitude = entity.isFlying() ? 20 : 0;
		this.lifePoints = lifePoints;
	}

	/**
	 * animate the warrior (move to destination and determine next sprite frame)
	 */
	public void animate() {
		super.animate();

		if (x == xd && y == yd && entity.getState() == 'r') {
			visible = false;
		}

		if (action == 'd') {
			if (entity.isFlying()) {
				if (animationCounter == 2) {
					setAltitude(10);
				} else if (animationCounter == 3) {
					setAltitude(10);
				}
			}

		}
	}

	/**
	 * events managment when a property changed
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		if (evt.getPropertyName().equals("scale")) {
			float previousScale = ((Float) evt.getOldValue()).floatValue();
			Rectangle previousBounds = getBounds(previousScale);
			firePropertyChange(BOUNDS, previousBounds, getBounds());
		} else if (evt.getPropertyName().equals(Entity.FLYING)) {
			int newAltitude = ((Boolean) evt.getNewValue()).booleanValue() ? 20
					: 0;
			setAltitude(newAltitude);
		}
	}

	/**
	 * change altitude
	 * 
	 * @param newAltitude
	 *            the new altitude
	 */
	public void setAltitude(int newAltitude) {
		if (newAltitude != altitude) {
			Rectangle previousBounds = getBounds();
			altitude = newAltitude;
			firePropertyChange(BOUNDS, previousBounds, getBounds());
		}
	}

	/**
	 * initialize contact areas
	 */
	public void init() {
		super.init();
		contactArea = new Area(new Rectangle(modelDx() - 2, modelDy() - 2,
				width() + 4, length() + 4));
		spearContactArea = new Area(new Rectangle(3 * modelDx() - 2,
				3 * modelDy() - 2, 3 * width() + 4, 3 * length() + 4));
		pikeContactArea = new Area(new Rectangle(5 * modelDx() - 2,
				5 * modelDy() - 2, 5 * width() + 4, 5 * length() + 4));
		pikeContactArea.subtract(spearContactArea);
		spearContactArea.subtract(contactArea);
	}

	/**
	 * reset contact areas
	 */
	public void resetAreas() {
		// create
		super.resetAreas();
		contactArea = new Area(new Rectangle(modelDx() - 2, modelDy() - 2,
				width() + 4, length() + 4));
		spearContactArea = new Area(new Rectangle(3 * modelDx() - 2,
				3 * modelDy() - 2, 3 * width() + 4, 3 * length() + 4));
		pikeContactArea = new Area(new Rectangle(5 * modelDx() - 2,
				5 * modelDy() - 2, 5 * width() + 4, 5 * length() + 4));
		pikeContactArea.subtract(spearContactArea);
		spearContactArea.subtract(contactArea);
		// move

		contactArea = new Area(AffineTransform.getTranslateInstance(
				(double) xd, (double) yd).createTransformedShape(contactArea));
		spearContactArea = new Area(AffineTransform.getTranslateInstance(
				(double) xd, (double) yd).createTransformedShape(
				spearContactArea));
		pikeContactArea = new Area(AffineTransform.getTranslateInstance(
				(double) xd, (double) yd).createTransformedShape(
				pikeContactArea));
		// rotate

		contactArea = new Area(AffineTransform.getRotateInstance(-angle,
				(double) xd, (double) yd).createTransformedShape(contactArea));
		spearContactArea = new Area(AffineTransform.getRotateInstance(-angle,
				(double) xd, (double) yd).createTransformedShape(
				spearContactArea));
		pikeContactArea = new Area(AffineTransform.getRotateInstance(-angle,
				(double) xd, (double) yd).createTransformedShape(
				pikeContactArea));
	}

	/**
	 * get the warrior id
	 * 
	 * @return the warrior id
	 */
	public int getId() {
		return BattleFieldModel.getInstance().getId(this);
	}

	/**
	 * remove specified number of lifepoints
	 * 
	 * @param wounds
	 *            number of life points to remove
	 */
	public void wound(int wounds) {
		if (FB.isNet()) {
			Client.getInstance().sendAndWait(
					MessagesConstants.WOUNDMODEL + getId() + "#" + wounds);
		} else {
			doWound(wounds);
		}
	}

	/**
	 * play dying animation
	 * 
	 */
	public void animKill() {
		mustDie = false;
		setAction('d');
		animationCounter = 1;
	}

	/**
	 * do remove specified number of lifepoints
	 * 
	 * @param wounds
	 *            number of life points to remove
	 */
	public void doWound(int wounds) {

		lifePoints -= wounds;
		// FB.debug(entity.getFullName()+" receives "+wounds+" wounds , remains
		// "+pdv );
		entity.firePropertyChange("pdv", lifePoints + wounds, lifePoints);
		if (lifePoints <= 0) {
			mustDie = true;
			entity.setDead(this);
		}
	}

	/**
	 * is warrior currently walking ?
	 * 
	 * @return true if the warrior is walking
	 */
	public boolean iswalking() {
		return action == 'w' && lifePoints > 0 && visible;
	}

	/**
	 * clone the warrior
	 */
	public Object clone() {
		return super.clone();
	}

	/**
	 * a possibility of surrounding a model
	 * 
	 * @author farfpeuf
	 * 
	 */

	public class SurroundingPossibility {
		// position
		public Point position;

		// orientation
		public double theta;

		/**
		 * constructor
		 * 
		 * @param oX
		 * @param oY
		 * @param theta
		 *            the orientation
		 * @param fig
		 *            the surrounded warrior
		 */
		SurroundingPossibility(int oX, int oY, double theta, Warrior fig) {
			Point o = new Point(oX, oY);
			o.translate(-fig.xd, -fig.yd);
			AffineTransform.getRotateInstance(-fig.angle).deltaTransform(o, o);
			o.translate(fig.xd, fig.yd);
			position = o;
			this.theta = theta;
		}

	}

	/**
	 * calculate the surrounding possibilities arround this warrior
	 * 
	 * @param otherWidth
	 *            ennemy base width
	 * @param otherlength
	 *            ennemy base length
	 * @return a list of SurroundingPossibility
	 */
	public List calculateSurroundinPossibilities(float otherWidth,
			float otherlength) {

		List possibilities = new ArrayList();

		int pasW = (int) (otherWidth / 2);
		int pasL = (int) (otherlength / 2);

		for (int xi = xd + modelDx() - pasW; xi < xd - modelDx() + pasW; xi += (pasW + 1)) {
			possibilities.add(new SurroundingPossibility(xi, yd - modelDy()
					+ pasL, angle, this));
			possibilities.add(new SurroundingPossibility(xi, yd + modelDy()
					- pasL, angle + Math.PI, this));
		}
		possibilities.add(new SurroundingPossibility(xd - modelDx() + pasW, yd
				- modelDy() + pasL, angle, this));
		possibilities.add(new SurroundingPossibility(xd - modelDx() + pasW, yd
				+ modelDy() - pasL, angle + Math.PI, this));

		for (int yi = yd + modelDy() - pasW; yi < yd - modelDy() + pasW; yi += (pasW + 1)) {
			possibilities.add(new SurroundingPossibility(xd + modelDx() - pasL,
					yi, angle - Math.PI / 2.0, this));
			possibilities.add(new SurroundingPossibility(xd - modelDx() + pasL,
					yi, angle + Math.PI / 2.0, this));
		}
		possibilities.add(new SurroundingPossibility(xd + modelDx() - pasL, yd
				- modelDy() + pasW, angle - Math.PI / 2.0, this));
		possibilities.add(new SurroundingPossibility(xd - modelDx() + pasL, yd
				- modelDy() + pasW, angle + Math.PI / 2.0, this));
		return possibilities;
	}

	/**
	 * play routing animation
	 * 
	 */
	public void rout() {
		Point startPoint = entity.getArmy().getStartPoint();
		int routDX = x - startPoint.x;
		int routDY = y - startPoint.y;
		double routDist = Math.sqrt(routDX * routDX + routDY * routDY);
		routDX = (int) (50f / routDist * routDX);
		routDY = (int) (50f / routDist * routDY);
		moveTo(x - routDX + DX.roll(30) - 15, y - routDY + DX.roll(30) - 15);
		animMovement();
	}

	/**
	 * calculate the warrior contacts
	 * 
	 */
	public void calculateContacts() {
		Warrior model2;
		Entity entity2;
		int nbmodels = BattleFieldModel.getInstance().getNbWarriors();
		int nature;
		Object element;
		for (int j = 0; j < nbmodels; j++) {
			element = BattleFieldModel.getInstance().getWarrior(j);
			if (element instanceof Warrior) {
				model2 = (Warrior) element;
				entity2 = model2.entity;
				if (entity2 != spriteDelegate) {
					nature = touches(model2);
					if (nature != -1) {
						addContact(model2, nature);
						model2.addContact(this, nature);
					}
				}
			}
		} // for
	}

	/**
	 * add a new contact with this warrior
	 * 
	 * @param model2
	 *            other warrior
	 * @param nature
	 *            type of contact
	 */
	public void addContact(Warrior model2, int nature) {
		if (nature == 0) {
			friendlycontacts.add(model2);
		} else if (nature == 1) {
			ennemyContacts.add(model2);
		} else if (nature == 2) {
			ennemySpearContacts.add(model2);
		} else {
			ennemyPikeContacts.add(model2);
		}
		entity.addContact(model2.entity, nature);
	}

	/**
	 * remove a contact with this warrior
	 * 
	 * @param model2
	 *            other warrior
	 * @param nature
	 *            type of contact
	 */
	public void removecontact(Warrior model2, int nature) {
		if (nature == 0) {
			friendlycontacts.remove(model2);
		} else if (nature == 1) {
			ennemyContacts.remove(model2);
		} else if (nature == 2) {
			ennemySpearContacts.remove(model2);
		} else {
			ennemyPikeContacts.remove(model2);
		}
		entity.removeContact(model2.entity, nature);
	}

	/**
	 * clear this warrior contact
	 * 
	 */
	public void clearcontacts() {
		for (int i = 0; i < friendlycontacts.size(); i++) {
			((Warrior) (friendlycontacts.get(i))).removecontact(this, 0);
		}
		for (int i = 0; i < ennemyContacts.size(); i++) {
			((Warrior) (ennemyContacts.get(i))).removecontact(this, 1);
		}
		for (int i = 0; i < ennemySpearContacts.size(); i++) {
			((Warrior) (ennemySpearContacts.get(i))).removecontact(this, 2);
		}
		for (int i = 0; i < ennemyPikeContacts.size(); i++) {
			((Warrior) (ennemyPikeContacts.get(i))).removecontact(this, 3);
		}
		friendlycontacts.clear();
		ennemyContacts.clear();
		ennemySpearContacts.clear();
		ennemyPikeContacts.clear();
	}

	/**
	 * does this warrior touches another warrior ?
	 * @param model the other warrior
	 * @return true if the 2 warriors are in base to base contact
	 */
	public int touches(Warrior model) {
		if (!entity.sameAltitude(model.entity)) {
			return -1;
		}
		Area inter;
		inter = (Area) contactArea.clone();
		inter.intersect(model.baseArea);
		if (!entity.isEnnemy(model.entity)) {
			if (!inter.isEmpty()) {
				return 0; // friendly
			}
		} else { // opposing
			if (!inter.isEmpty()) {
				return 1; // first row
			}
			inter = (Area) spearContactArea.clone();
			inter.intersect(model.baseArea);
			if (!inter.isEmpty()) {
				return 2; // 2nd row
			}
			inter = (Area) pikeContactArea.clone();
			inter.intersect(model.baseArea);
			if (!inter.isEmpty()) {
				return 3; // 3rd row
			}
		}
		return -1;
	}

	/**
	 * move this warrior to a new destination
	 */
	public void moveTo(int nx, int ny) {
		int xt = nx - xd, yt = ny - yd;

		contactArea = new Area(AffineTransform.getTranslateInstance(
				(double) xt, (double) yt).createTransformedShape(contactArea));
		spearContactArea = new Area(AffineTransform.getTranslateInstance(
				(double) xt, (double) yt).createTransformedShape(
				spearContactArea));
		pikeContactArea = new Area(AffineTransform.getTranslateInstance(
				(double) xt, (double) yt).createTransformedShape(
				pikeContactArea));
		super.moveTo(nx, ny);
	}

	/**
	 * change warrior orientation
	 */
	public void setAngle(double newAngle) {
		rotate(newAngle - angle);
	}

	/**
	 * rotate warrior
	 */
	public void rotate(double deltaAngle) {

		contactArea = new Area(AffineTransform.getRotateInstance(-deltaAngle,
				(double) xd, (double) yd).createTransformedShape(contactArea));
		spearContactArea = new Area(AffineTransform.getRotateInstance(
				-deltaAngle, (double) xd, (double) yd).createTransformedShape(
				spearContactArea));
		pikeContactArea = new Area(AffineTransform.getRotateInstance(
				-deltaAngle, (double) xd, (double) yd).createTransformedShape(
				pikeContactArea));
		super.rotate(deltaAngle);
	}

	/**
	 * paint warrior contact area on the battlefield
	 * @param g graphics
	 * @param dx x offset
	 * @param dy y offset
	 * @param scale view scale
	 */
	public void paintContactArea(Graphics g, int dx, int dy, int scale) {
		Color c = g.getColor();
		g.setColor(Color.pink);
		Graphics2D g2d = (Graphics2D) g;
		Area scalecontact = new Area(AffineTransform.getScaleInstance(scale,
				scale).createTransformedShape(contactArea));
		scalecontact = new Area(AffineTransform.getTranslateInstance(dx, dy)
				.createTransformedShape(scalecontact));
		g2d.draw(scalecontact);
		g.setColor(c);
	}

	/**
	 * get the warrior bounds
	 */
	public Rectangle getBounds(float entityScale) {

		if (FB.getInstance() == null || FB.getInstance().isBatch()) {
			return new Rectangle();
		}

		int imageWidth;
		int imageHeight;

		if (altitude > 0) {
			imageWidth = ((FlyingEntity) spriteDelegate).getFlyingSprite()
					.getFrameWitdh();
			imageHeight = ((FlyingEntity) spriteDelegate).getFlyingSprite()
					.getFrameHeight();
		} else {
			imageWidth = spriteDelegate.getSprite().getFrameWitdh();
			imageHeight = spriteDelegate.getSprite().getFrameHeight();
		}

		int rx = (int) (x - imageWidth * entityScale / 2);
		int ry = (int) (y - altitude - imageHeight * entityScale / 2);
		int rw = (int) (imageWidth * entityScale);
		int rh = (int) (imageHeight + altitude * entityScale);
		return new Rectangle(rx, ry, rw, rh);
	}

	/**
	 * paint the warrior image on the battlefield
	 */
	public void paintImage(Graphics g, int dx, int dy, int scale) {
		if (visible) {

			if (entity.isFlying()) {

				Composite old_compo = ((Graphics2D) g).getComposite();
				((Graphics2D) g).setComposite(AlphaComposite.getInstance(
						AlphaComposite.SRC_OVER, 0.5f));
				((FlyingEntity) spriteDelegate).getShadowSprite().drawFrame(g,
						x, y, scale, dx, dy, frameIndex,
						spriteDelegate.getScale());
				((Graphics2D) g).setComposite(old_compo);
				((FlyingEntity) spriteDelegate).getFlyingSprite().drawFrame(g,
						x, y - altitude, scale, dx, dy, frameIndex,
						spriteDelegate.getScale());
			} else {
				super.paintImage(g, dx, dy, scale);
			}

		} // visible
	}

	/**
	 * play warrior fighting animation
	 *
	 */
	public void animfight() {
		action = 'f';
		double angledecombat = 0.0, mindist = 9999.9, disto;
		Warrior model;
		// get the closest opponent
		for (int i = 0; i < ennemyContacts.size(); i++) {
			model = (Warrior) (ennemyContacts.get(i));
			disto = distanceto(model);
			if (disto < mindist) {
				mindist = disto;
				angledecombat = angleTo(model);
			}
		}
		if (ennemyContacts.size() == 0) {
			for (int i = 0; i < ennemySpearContacts.size(); i++) {
				model = (Warrior) (ennemySpearContacts.get(i));
				disto = distanceto(model);
				if (disto < mindist) {
					mindist = disto;
					angledecombat = angleTo(model);
				}
			}
			if (ennemySpearContacts.size() == 0) {
				for (int i = 0; i < ennemyPikeContacts.size(); i++) {
					model = (Warrior) (ennemyPikeContacts.get(i));
					disto = distanceto(model);
					if (disto < mindist) {
						mindist = disto;
						angledecombat = angleTo(model);
					}
				}
			}
		}
		setCurrentSpritsequence(GeometryTools.to8Base(angledecombat));
	}

	/**
	 * play warrior threating animation
	 * @param threatingAngle angle to target
	 */
	public void animthreat(double threatingAngle) {
		action = 't';
		animationCounter = -DX.roll(4);
		tmpAngle = threatingAngle;
	}

	/**
	 * calculate distance to another warrior
	 * @param model the other warrior
	 * @return the distance
	 */
	public double distanceto(Warrior model) {
		int X = model.x - x;
		int Y = model.y - y;
		return Math.sqrt(X * X + Y * Y);
	}
	/**
	 * calculate angle to another warrior
	 * @param model the other warrior
	 * @return the angle in radians
	 */
	public double angleTo(Warrior model) {
		int X = model.x - x;
		int Y = model.y - y;
		double deupi = 2 * Math.PI, theta;
		theta = -Math.atan2(Y, X);
		theta -= Math.PI / 2;
		return theta % deupi;
	}
/**
 * is this warrior currently flying ?
 * @return true if the warrior is flying
 */
	public boolean isflying() {
		return entity.isFlying();
	}

	/**
	 * get entity to which this warrior belongs 
	 * @return the entity to which this warrior belongs 
	 */
	public Entity getEntity() {
		return entity;
	}

	/**
	 * @return Returns the lifePoints.
	 */
	public int getLifePoints() {
		return lifePoints;
	}

	/**
	 * @param lifePoints
	 *            The lifePoints to set.
	 */
	public void setLifePoints(int lifePoints) {
		this.lifePoints = lifePoints;
	}

	/**
	 * @return Returns the mustdie.
	 */
	public boolean mustDie() {
		return mustDie;
	}

	/**
	 * @return Returns the contactArea.
	 */
	public Area getContactArea() {
		return contactArea;
	}

	/**
	 * @return Returns the pikeContactArea.
	 */
	public Area getPikeContactArea() {
		return pikeContactArea;
	}

	/**
	 * @return Returns the spearContactArea.
	 */
	public Area getSpearContactArea() {
		return spearContactArea;
	}

	/**
	 * @return Returns the mustDie.
	 */
	public boolean isMustDie() {
		return mustDie;
	}

	/**
	 * @param mustDie
	 *            The mustDie to set.
	 */
	public void setMustDie(boolean mustDie) {
		this.mustDie = mustDie;
	}

	/**
	 * heal the warrior
	 * 
	 * @param lifePoints
	 *            added life points
	 */
	public void heal(int lifePoints) {
		this.lifePoints += lifePoints;
	}

	/**
	 * @return Returns the opposingcontacts.
	 */
	public List getEnnemyContacts() {
		return ennemyContacts;
	}

	/**
	 * @return Returns the secondcontacts.
	 */
	public List getEnnemySpearContacts() {
		return ennemySpearContacts;
	}

	/**
	 * @return Returns the thirdcontacts.
	 */
	public List getEnnemyPikeContacts() {
		return ennemyPikeContacts;
	}
}