/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.campaign;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.Area;
import java.util.List;

import fb.army.hierarchy.Force;
import fb.gui.panels.GamePanel;
import fb.main.TimerListener;

/**
 * game panel where hexagonal map is painted
 * 
 * @author farfpeuf
 * 
 */
public class HexagonalMapPanel extends GamePanel implements TimerListener {

	// the map painting delegate
	protected HexagonalMapPainter mapPainter;

	// current tile uneder mouse cursor
	private Point currentTileCoord = new Point();

	/**
	 * constructor build panel and register mouse motions
	 * 
	 */
	public HexagonalMapPanel() {
		super();
		mapPainter = new HexagonalMapPainter();
		addMouseMotionListener(new MouseMotionAdapter() {
			public void mouseMoved(MouseEvent e) {
				Point old = getCurrentTileCoord();
				Point newP = getHexaCoord(e.getPoint());

				if (!newP.equals(old)) {
					setCurrentTileCoord(newP);
					refresh(getHexaRectangle(old));
					refresh(getHexaRectangle(newP));
				}
			}
		});
	}

	/**
	 * center view on specified point
	 * @param hexaCoord point hexagonal coordinates
	 */
	public void centerOn(Point hexaCoord) {
		Point maxiMouse = getMouseCoord(hexaCoord);

		int screenCenterX = getWidth() / 2;
		int screenCenterY = getHeight() / 2;
		setDx(screenCenterX - maxiMouse.x);
		setDy(screenCenterY - maxiMouse.y);
		refreshAll();
	}

	/**
	 * get tile rectangular bounds
	 * @param hexaCoord the tile coordinates
	 * @return the rectangular bounds
	 */
	public Rectangle getHexaRectangle(Point hexaCoord) {
		Rectangle result = new Rectangle();
		Point center = getTranslatedMouseCoord(hexaCoord);

		result.x = center.x - getRadius() - 1;
		result.y = (int) (center.y - HexagonalMapPainter.SIN_PI_SUR_3
				* getRadius()) - 1;
		result.width = getRadius() * 2 + 2;
		result.height = (int) (2 * HexagonalMapPainter.SIN_PI_SUR_3 * getRadius()) + 2;

		return result;
	}

	/**
	 * set current highlighted tile
	 * @param currentTileCoord the new tile coordinates
	 */
	public void setCurrentTileCoord(Point currentTileCoord) {

		this.currentTileCoord = currentTileCoord;
	}

	/**
	 * get current highlighted tile
	 * @return the tile coordinates
	 */
	public Point getCurrentTileCoord() {
		return currentTileCoord;
	}

	/**
	 * paint all tiles highlights
	 * @param g graphics
	 */
	public void paintHighLight(Graphics g) {

		Composite old_compo = ((Graphics2D) g).getComposite();

		((Graphics2D) g).setComposite(AlphaComposite.getInstance(
				AlphaComposite.SRC_OVER, 0.3f));
		g.setColor(Color.white);

		drawTileHighLight(g, currentTileCoord);
		((Graphics2D) g).setComposite(AlphaComposite.getInstance(
				AlphaComposite.SRC_OVER, 0.5f));
		List mvtZone = Campaign.getInstance().getMovementZone();
		if (mvtZone != null) {
			for (int i = 0; i < mvtZone.size(); i++) {
				drawTileHighLight(g, (Point) mvtZone.get(i));
			}

		}

		((Graphics2D) g).setComposite(old_compo);

	}

	/**
	 * paint single tile highlight
	 * @param g graphics
	 * @param tileCoord the tile coordinates
	 */
	public void drawTileHighLight(Graphics g, Point tileCoord) {
		Shape oldClip = g.getClip();
		if (tileCoord.x >= 0 && tileCoord.y >= 0
				&& tileCoord.x < HexagonalMapModel.getInstance().getMapWidth()
				&& tileCoord.y < HexagonalMapModel.getInstance().getMapLength()) {

			Point center = getMouseCoord(tileCoord);
			Area clip = new Area(new Hexagon(center.x, center.y, getRadius()));
			clip.intersect(new Area(oldClip));
			g.setClip(clip);
			g.fillRect(center.x - getRadius(), center.y - getRadius(),
					getRadius() * 2, getRadius() * 2);
		}
		g.setClip(oldClip);
	}

	/**
	 * convert mouse coordinates to hexagonal (row,column) coordinates
	 * @param mouseCoord the mouse coordinates
	 * @return the corresponding hexagonal coordinates
	 */
	public Point getHexaCoord(Point mouseCoord) {
		Point dMouse = new Point(mouseCoord.x - dx, mouseCoord.y - dy);
		return mapPainter.getHexaCoord(dMouse);
	}

	/**
	 * convert hexagonal coordinates (row,column) to translated (view x and y offset) mouse coordinates
	 * @param hexaCoord the hexagonal coordinates
	 * @return the corresponding translated mouse coordinates
	 */
	public Point getTranslatedMouseCoord(Point hexaCoordCoord) {
		Point mouseCoord = getMouseCoord(hexaCoordCoord);
		mouseCoord.translate(dx, dy);
		return mouseCoord;
	}

	/**
	 * convert hexagonal coordinates (row,column) to mouse coordinates
	 * @param hexaCoord the hexagonal coordinates
	 * @return the corresponding mouse coordinates
	 */
	public Point getMouseCoord(Point hexaCoordCoord) {
		return mapPainter.getMouseCoord(hexaCoordCoord);
	}

	/**
	 * get hexagons radius
	 * @return hexagons radius
	 */
	public int getRadius() {
		return mapPainter.getRadius();
	}

	/**
	 * change hexagons radius (zoom in and out)
	 * @param i the new radius
	 */
	public synchronized void setRadius(int i) {
		mapPainter.setRadius(i);
		refreshAll();
	}

	/**
	 * draw the whole panel
	 */
	protected void draw(Graphics g) {
		g.translate(dx, dy);
		mapPainter.paintMap(g);
		Force currentForce = Campaign.getInstance().currentForce();
		if (currentForce != null && !currentForce.isAi()
				&& currentForce.isHandled()) {
			paintHighLight(g);
		}
		g.translate(-dx, -dy);

	}

	/**
	 * change the map data model
	 * @param model the new map data model
	 */
	public void setModel(HexagonalMapModel model) {
		mapPainter.setModel(model);
	}

}
