/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.gui.widgets.armytreetable;

import javax.swing.JTree;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.tree.TreePath;

import fb.army.hierarchy.HierarchyNode;

/**
 * a table to tree and tree to table model adapter
 * 
 * @author farfpeuf
 */

public class TreeTableModelAdapter extends AbstractTableModel {
	/**
	 * tree delegate
	 */
	private JTree tree;

	/**
	 * tree table data model
	 */
	private TreeTableModel treeTableModel;

	/**
	 * constructor
	 * 
	 * @param treeTableModel
	 *            the tree table data model
	 * @param tree
	 *            the tree delegate
	 */
	public TreeTableModelAdapter(TreeTableModel treeTableModel, JTree tree) {
		this.tree = tree;
		this.treeTableModel = treeTableModel;

		tree.addTreeExpansionListener(new TreeExpansionListener() {
			// Don't use fireTableRowsInserted() here;
			// the selection model would get updated twice.
			public void treeExpanded(TreeExpansionEvent event) {
				fireTableDataChanged();
			}

			public void treeCollapsed(TreeExpansionEvent event) {
				fireTableDataChanged();
			}
		});

		tree.getModel().addTreeModelListener(new TreeModelListener() {
			public void treeNodesChanged(TreeModelEvent e) {
				Object[] nodes = e.getChildren();
				HierarchyNode node;
				int row;
				if (nodes != null) {
					for (int i = 0; i < nodes.length; i++) {
						node = (HierarchyNode) nodes[i];
						row = rowForNode(node);
						fireTableRowsUpdated(row, row);
					}
				}
			}

			public void treeNodesInserted(TreeModelEvent e) {
				fireTableStructureChanged();
			}

			public void treeNodesRemoved(TreeModelEvent e) {
				fireTableStructureChanged();
			}

			public void treeStructureChanged(TreeModelEvent e) {
				fireTableStructureChanged();
			}
		});

	}

	/**
	 * Returns the number ofs available column.
	 */
	public int getColumnCount() {
		return treeTableModel.getColumnCount();
	}

	/**
	 * Returns the name for column number <code>column</code>.
	 */
	public String getColumnName(int column) {
		return treeTableModel.getColumnName(column);
	}

	/**
	 * Returns the type for column number <code>column</code>.
	 */
	public Class getColumnClass(int column) {
		return treeTableModel.getColumnClass(column);
	}

	/**
	 * Returns the number ofs available rows.
	 */
	public int getRowCount() {
		return tree.getRowCount();
	}

	/**
	 * get tree node corresponding to row
	 * 
	 * @param row
	 *            row number
	 * @return corresponding node
	 */
	public Object nodeForRow(int row) {
		TreePath treePath = tree.getPathForRow(row);
		if (treePath == null) {
			return null;
		}
		return treePath.getLastPathComponent();
	}

	/**
	 * get row number corresponding to tree node
	 * 
	 * @param node
	 *            the tree node
	 * @return corresponding row number
	 */
	public int rowForNode(Object node) {
		for (int i = 0; i < getRowCount(); i++) {
			if (node == nodeForRow(i)) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Returns the value to be displayed for node <code>node</code>, at
	 * column number <code>column</code>.
	 */
	public Object getValueAt(int row, int column) {
		return treeTableModel.getValueAt(nodeForRow(row), column);
	}

	/**
	 * Indicates whether the the value for node <code>node</code>, at column
	 * number <code>column</code> is editable.
	 */
	public boolean isCellEditable(int row, int column) {
		return treeTableModel.isCellEditable(nodeForRow(row), column);
	}

	/**
	 * Sets the value for node <code>node</code>, at column number
	 * <code>column</code>.
	 */
	public void setValueAt(Object value, int row, int column) {
		treeTableModel.setValueAt(value, nodeForRow(row), column);
	}

	/**
	 * @return the tree table model
	 */
	public TreeTableModel getTreeTableModel() {
		return treeTableModel;
	}

}
